from dataclasses import dataclass, field
from dotenv import dotenv_values
from typing import *

@dataclass
class SystemSettings:
    openai_api_key: str

    def __post_init__(self):
        self.openai_api_key = str(self.openai_api_key)


def load_env_vars() -> SystemSettings:
    env_vars = dict(dotenv_values())
    new = {}
    for index, (key, value) in enumerate(env_vars.items()):
        new[key.lower()] = value
    lower_case_settings = dict(map(lambda x: (x[0].lower(), x[1]), env_vars.items()))
    return SystemSettings(**lower_case_settings)


from __future__ import annotations
from json import JSONDecodeError

import openai
import json
from typing import *
from dataclasses import dataclass, field
from .schema import GPTContext, write_schema
from pprint import pprint

class ContextClient:
    def __init__(self, api_key: str, task_text: str, rules: Rules = None):
        self.api_key = api_key
        self.task_text = task_text
        self.rules = rules
    def create_task(self, inputs: GPTContext | List[GPTContext], output: Type[GPTContext], object_context: bool = False) -> dict:
        result = create_task(self.api_key, self.task_text, inputs, output, rules=self.rules, object_context=object_context)
        return result


class InvalidJson(Exception):
    def __init__(self, data: str, message="Unable to process json returned by gpt"):
        self.data = data
        self.message = message
        super().__init__(self.message)

@dataclass
class Rules:
    # main rules
    follow_input_and_output: bool = field(default=True)
    output_must_be_json: bool = field(default=True)
    output_must_follow_schema: bool = field(default=True)

    treat_numbers_as_currency: bool = field(default=False)

    others: List[str] = field(default_factory=list)

    @property
    def flatten(self) -> str:
        rules = ["All inputs will be valid json"]
        if self.follow_input_and_output:
            rules.append("Follow the Input and Output formats provided")
        if self.output_must_be_json:
            rules.append("All outputs must be valid json")
        if self.output_must_follow_schema:
            rules.append("All outputs must follow the schema provided")

        if self.treat_numbers_as_currency:
            rules.append("Treat all integers as USD currency")

        for rule in self.others:
            rules.append(rule)
        return "\n- ".join(rules)


def create_task(api_key: str, task: str, input_obj: GPTContext | List[GPTContext], output_obj: Type[GPTContext],
                rules: Rules = None, object_context: bool = False, debug: bool = False) -> dict | None:
    openai.api_key = api_key
    if rules is None:
        rules = Rules()
    if isinstance(input_obj, list):
        temp_input_obj = input_obj[0]
        input_data = []
        for item in input_obj:
            input_data.append(item.flatten)
    else:
        temp_input_obj = input_obj
        input_data = input_obj.flatten

    if object_context:
        context_str = f"Input Object Name: {temp_input_obj.__class__.__name__}\nOutput Object Name: {output_obj.__name__}"
    else:
        context_str = ""

    messages = [
        {
            "role": "system",
            "content": "You are a program that completes the following task and responds with valid json.\n\n"
                       ""
                       f"Task:\n{task}\n\n\n"
                       f"Rules:\n- {rules.flatten}\n\nInput Format:\n{str(temp_input_obj.schema)}\n\nOutput Format:\n{str(write_schema(output_obj))}\n\n{context_str}\n\n"
        },
        {"role": "user", "content": str(input_data)}
    ]
    if debug:
        print("Messages:\n")
        pprint(messages)
        print("\n")
    # print(messages[0]["content"])
    # print(messages[1]["content"])
    # return
    response = openai.ChatCompletion.create(
        model="gpt-4",
        messages=messages,
        temperature=1,
        max_tokens=1000,
        top_p=1,
        frequency_penalty=0,
        presence_penalty=0,
        # stop=[".............."]

    )
    try:
        loaded_resp = dict(response["choices"][0]["message"]["content"])
    except ValueError:
        try:
            loaded_resp = json.loads(response["choices"][0]["message"]["content"])
        except JSONDecodeError:
            print(repr(response["choices"][0]["message"]["content"]))
            raise InvalidJson(response["choices"][0]["message"]["content"])

    return loaded_resp



from enviroment import load_env_vars
from typing import *
from dataclasses import dataclass
from GPTContext.GPTContext import GPTContext, Rules, create_task


env_vars = load_env_vars()


@dataclass
class Toy(GPTContext):
    color: str
    price: int


@dataclass
class Dog(GPTContext):
    name: str
    age: int

    favorite_toy: Toy
    favorite_toys: List[Toy]


@dataclass
class DogDesc(GPTContext):
    name: str
    description: str


def test_single():
    bob_toys = [Toy("red", 5), Toy("blue", 6), Toy("green", 7)]

    bob = Dog("Bob", 5, bob_toys[1], bob_toys)
    rules = Rules(treat_numbers_as_currency=True)
    task = "write a description for each pet based on the provided characteristics"
    result = create_task(env_vars.openai_api_key, task, bob, DogDesc, rules=rules)
    print(DogDesc(**result))


def test_multi():
    bob_toys = [Toy("red", 5), Toy("blue", 6), Toy("green", 7)]
    jade_toys = [Toy("pink", 10), Toy("orange", 600), Toy("yellow", 799)]

    bob = Dog("Bob", 5, bob_toys[1], bob_toys)
    jade = Dog("Jade", 14, jade_toys[1], jade_toys)

    rules = Rules(treat_numbers_as_currency=True)
    task = "write a description for each pet based on the provided characteristics"
    result = create_task(env_vars.openai_api_key, task, [bob, jade], DogDesc, rules=rules, object_context=True)
    for res in result:
        print(DogDesc(**res))


test_multi()



from enviroment import load_env_vars
from dataclasses import dataclass
from GPTContext.GPTContext import GPTContext, Rules, create_task

env_vars = load_env_vars()


@dataclass
class Author(GPTContext):
    name: str


@dataclass
class Book(GPTContext):
    title: str
    author: Author

    description: str


@dataclass
class Review(GPTContext):
    book_title: str
    review: str


author = Author("Angie Thomas")
book = Book("The Hate U Give", author, "Sixteen-year-old Starr Carter moves between two worlds: the poor neighborhood where she lives and the fancy suburban prep school she attends. The uneasy balance between these worlds is shattered when Starr witnesses the fatal shooting of her childhood best friend Khalil at the hands of a police officer.")
rules = Rules(others=["limit response to 20 words"])
task = "Write a negative review about the book. Add heavy criticism towards the author"

result = create_task(env_vars.openai_api_key, task, book, Review, rules, object_context=True)
print(Review(**result))


from typing import *
import types
import builtins
from datetime import datetime

builtin_types = [getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)]
builtin_types += [datetime]


def write_schema(input_data: object) -> dict:
    output = {}
    type_hints = get_type_hints(input_data)
    # type_hints = input_data.__annotations__
    for index, (key, value) in enumerate(type_hints.items()):



        # print("\n")
        str_val = str(value)
        try:
            issubclass(value, str)
        except TypeError:
            value = type(value)
        # print(str_val)
        # print(type(value))
        type_indicator = None
        if str_val.startswith("typing."):
            if str_val.startswith("typing.List"):
                # get the type hints inside the List type hint
                list_args = type_hints.get(key).__args__
                # make sure that the object is a subclass
                if issubclass(list_args[0], GPTContext):
                    class_dict = write_schema(list_args[0])
                    type_indicator = f"List[{class_dict}]"
                else:
                    type_indicator = str_val.split(".")[1]
            elif str_val.startswith("typing.Tuple"):
                tuple_args = type_hints.get(key).__args__
                temp = []
                for arg in tuple_args:
                    if issubclass(arg, GPTContext):
                        class_dict = write_schema(arg)
                        temp.append(f"{arg.__name__}({str(class_dict)})")
                    else:
                        temp.append(str(arg).split("'")[1])
                type_indicator = f"Tuple[{', '.join(temp)}]"
            elif str_val.startswith("typing.Dict"):
                dict_args = type_hints.get(key).__args__
                temp = []
                for arg in dict_args:
                    if issubclass(arg, GPTContext):
                        class_dict = write_schema(arg)
                        temp.append(f"{arg.__name__}({str(class_dict)})")
                    else:
                        temp.append(str(arg).split("'")[1])
                type_indicator = f"Dict[{', '.join(temp)}]"
        elif issubclass(value, GPTContext):
            type_indicator = write_schema(value)

        elif value is types.UnionType:
            union_args = type_hints.get(key).__args__
            temp = []
            for arg in union_args:
                if issubclass(arg, GPTContext):
                    class_dict = write_schema(arg)
                    temp.append(f"{arg.__name__}({str(class_dict)})")
                else:
                    temp.append(str(arg).split("'")[1])
            type_indicator = f"Union[{', '.join(temp)}]"
        else:
            type_indicator = value.__name__


        output[key] = type_indicator
    return output


class GPTContext:

    @property
    def schema(self) -> dict:
        return write_schema(self)

    @property
    def flatten(self) -> dict:
        flattened = {}
        for attr_name in self.__annotations__.keys():
            current_val = self.__getattribute__(attr_name)
            if type(current_val) not in builtin_types and issubclass(type(current_val), GPTContext):
                flattened_val = current_val.flatten
            elif isinstance(current_val, list) or isinstance(current_val, tuple):
                flattened_val = []
                for item in current_val:
                    if issubclass(type(item), GPTContext):
                        flattened_val.append(item.flatten)
                    else:
                        flattened_val.append(item)
                if isinstance(current_val, tuple):
                    flattened_val = tuple(flattened_val)
            else:
                flattened_val = current_val
            flattened[attr_name] = flattened_val
        return flattened
